import 'package:flutter_inapp_purchase/flutter_inapp_purchase.dart';
import 'package:intl/intl.dart';
import 'package:knt_bloc/knt_bloc.dart';
import 'package:purchases_flutter/purchases_flutter.dart';

import '../model/subscription_item.dart';
import '../repository/monetized_repo.dart';
import '../service/revcat_iap_service.dart';

class RevCatSubscriptionBloc extends BaseBloc {
  final RevCatIapService _iapService;
  final MonetizedRepo _monetizedRepo;

  // In-app-purchase service
  late bool _purchased;

  RevCatSubscriptionBloc(this._iapService, this._monetizedRepo) {
    _initData();
    add(ConnectOldSubscriptionServiceEvent());
  }

  @override
  Stream<BaseState> handleEvent(BaseEvent event) async* {
    yield* super.handleEvent(event);
    switch (event.runtimeType) {
      case ConnectOldSubscriptionServiceEvent:
        yield* _handleConnectSubscriptionServiceEvent(
            event as ConnectOldSubscriptionServiceEvent);
        break;

      case RCCheckEntitlementEvent:
        yield* _handleCheckEntitlementEvent(event as RCCheckEntitlementEvent);
        break;

      case RCRequestSubscriptionEvent:
        yield* _handleRequestSubscriptionEvent(
            event as RCRequestSubscriptionEvent);
        break;

      case RCRestoreSubscriptionEvent:
        yield* _handleRestoreSubscriptionEvent(
            event as RCRestoreSubscriptionEvent);
        break;

      case RCFetchOfferingsEvent:
        yield* _handleFetchOfferingsEvent(event as RCFetchOfferingsEvent);
        break;

      case _RCUpdatePurchaseInfoEvent:
        yield* _handleUpdatePurchaseInfoEvent(
            event as _RCUpdatePurchaseInfoEvent);
        break;
    }
  }

  @override
  Stream<BaseState> handleBaseErrorEvent(BaseErrorEvent event) async* {
    if (event.tag == (RCFetchOfferingsEvent).toString()) {
      yield RCFetchedOfferingsFailureState();
      return;
    }
    if (event.tag == (RCRestoreSubscriptionEvent).toString()) {
      yield RCFetchedRestoreFailureState();
      return;
    }
    if (event.tag == (RCRequestSubscriptionEvent).toString()) {
      yield RCUnPurchasedState();
      return;
    }
    yield* super.handleBaseErrorEvent(event);
  }

  Future<void> _initData() async {
    _purchased = false;
  }

  Stream<BaseState> _handleConnectSubscriptionServiceEvent(
      ConnectOldSubscriptionServiceEvent event) async* {
    // Make listener at first
    Purchases.addPurchaserInfoUpdateListener((purchaserInfo) async {
      // Receives an updated purchaser info object from calls to getPurchaserInfo(),
      // purchasePackage(), purchaserProduct(), or restoreTransactions()
      _updatePurchaserInfo(purchaserInfo);
    });
    // Get subscriber info from RevenueCat
    final purchaserInfo = await Purchases.getPurchaserInfo();
    final subscriberFirstSeenDate = purchaserInfo.firstSeen.parse();
    final requestDate = purchaserInfo.requestDate.parse();
    final userNewToRevenueCat =
        subscriberFirstSeenDate.isSameDayTillHour(requestDate);
    final hasEntitlement =
        purchaserInfo.entitlements.all[_iapService.entitlementID]?.isActive ??
            false;
    if (hasEntitlement) {
      _purchased = true;
      add(_RCUpdatePurchaseInfoEvent());
      return;
    }
    if (!userNewToRevenueCat) {
      _purchased = false;
      add(_RCUpdatePurchaseInfoEvent());
      return;
    }

    // Request old system to check available SKUs, then sync to RevenueCat if
    // user has subscription
    await FlutterInappPurchase.instance.initConnection.then((value) async {
      await cancelableFuture(
          FlutterInappPurchase.instance.getAvailablePurchases());
      await FlutterInappPurchase.instance.endConnection;
    }).onError((error, stackTrace) {
      _purchased = false;
      add(_RCUpdatePurchaseInfoEvent());
    });
  }

  Stream<BaseState> _handleCheckEntitlementEvent(
      RCCheckEntitlementEvent event) async* {
    final purchaserInfo = await Purchases.getPurchaserInfo();
    _updatePurchaserInfo(purchaserInfo);
  }

  Stream<BaseState> _handleRequestSubscriptionEvent(
      RCRequestSubscriptionEvent event) async* {
    yield RCPurchasingState();
    final offerings = await Purchases.getOfferings();
    for (final package in offerings.current?.availablePackages ?? <Package>[]) {
      if (package.product.identifier == event.subItem.sku) {
        await Purchases.purchasePackage(package);
        return;
      }
    }
  }

  Stream<BaseState> _handleRestoreSubscriptionEvent(
      RCRestoreSubscriptionEvent event) async* {
    yield RCRestoringState();
    final purchaserInfo = await Purchases.restoreTransactions();
    _updatePurchaserInfo(purchaserInfo);
  }

  Stream<BaseState> _handleFetchOfferingsEvent(
      RCFetchOfferingsEvent event) async* {
    yield RCFetchingOfferingsState();
    final offerings = await Purchases.getOfferings();
    final subList = _iapService
        .packageConverter(offerings.current?.availablePackages ?? []);
    if (subList.isEmpty) {
      throw KntException(KntIssue.undefined);
    }
    yield RCFetchedOfferingsState(subList);
  }

  Stream<BaseState> _handleUpdatePurchaseInfoEvent(
      _RCUpdatePurchaseInfoEvent event) async* {
    await _monetizedRepo.savePurchasedIapStatus(purchased: _purchased);
    if (_purchased) {
      yield RCPurchasedState();
      return;
    }
    yield RCUnPurchasedState();
  }

  void _updatePurchaserInfo(PurchaserInfo purchaserInfo) {
    _purchased =
        purchaserInfo.entitlements.all[_iapService.entitlementID]?.isActive ??
            false;
    add(_RCUpdatePurchaseInfoEvent());
  }
}

// region EVENT
class ConnectOldSubscriptionServiceEvent extends BaseEvent {}

class RCCheckEntitlementEvent extends BaseEvent {}

class RCRequestSubscriptionEvent extends BaseEvent {
  final SubscriptionItem subItem;

  const RCRequestSubscriptionEvent(this.subItem);

  @override
  String toString() =>
      'RCRequestSubscriptionEvent: item has sku ${subItem.sku}';
}

class RCRestoreSubscriptionEvent extends BaseEvent {}

class RCFetchOfferingsEvent extends BaseEvent {}

class _RCUpdatePurchaseInfoEvent extends BaseEvent {}
// endregion

// region STATE
class RCPurchasingState extends BaseState {}

class RCRestoringState extends BaseState {}

class RCPurchasedState extends BaseState {}

class RCUnPurchasedState extends BaseState {}

class RCFetchingOfferingsState extends BaseState {}

class RCFetchedOfferingsState
    extends BaseHasResultState<List<SubscriptionItem>> {
  const RCFetchedOfferingsState(List<SubscriptionItem> subList)
      : super(subList);
}

class RCFetchedOfferingsFailureState extends BaseErrorState {
  RCFetchedOfferingsFailureState()
      : super(exception: KntException(KntIssue.undefined));
}

class RCFetchedRestoreFailureState extends BaseState {}
// endregion

extension DateUtils on DateTime {
  bool isSameDayTillHour(DateTime? date) {
    if (date == null) {
      return false;
    }
    return day == date.day &&
        month == date.month &&
        year == date.year &&
        hour == date.hour;
  }
}

extension InversionDate on String {
  static const revenueCatDateTimePattern = 'yyyy-MM-dd\'T\'HH:mm:ss\'Z\'';
  static const backupRevenueCatDateTimePattern =
      'yyyy-MM-dd\'T\'HH:mm:ss.SSS\'Z\'';

  DateTime parse() {
    try {
      return DateFormat(revenueCatDateTimePattern).parse(this);
    } on Exception {
      return DateFormat(backupRevenueCatDateTimePattern).parse(this);
    }
  }
}
